# -*- perl -*-

# Copyright (c) 1999 Jeff Weisberg <jaw+smt@op.net>
# see the file "License"

package SNMPTRAPInput;
@ISA = qw(Base);

use BER;
use Socket;
use SNMP_Session;
use Fcntl;
use Socket;

sub new {
    my $class = shift;
    my($files) = @_;
    my($self) = {};
    my $port = $Config::snmp_trapport || 162;
    
    return if $port eq "no";
    socket(::TRAP, PF_INET, SOCK_DGRAM, getprotobyname('udp')) || die "trap socket: $!";
    bind(::TRAP, sockaddr_in($port, INADDR_ANY))               || die "trap bind (port udp/$port): $!";

    bless $self, $class;
    $self->{fd} = ::TRAP;
    $self->set_nbio(1);
    
    $files->[ fileno(::TRAP) ] = $self;

    print STDERR "snmp: file ", fileno(::TRAP), ", port udp/$port\n";
    return $self;
}

@gentrap = qw(coldstart warmstart linkdown linkup authfail egbloss);

# handle trap input
sub readable {
    my $self = shift;
    my($from, $ip, $port, $buf, $e, $ifno);

    $from = recv( $self->{fd}, $buf="", 8192, 0);
    if( ! $from ){
	::error('WARN', "empty snmp/read");
	return -1;
    }
    ($port, $ip) = sockaddr_in($from);

    $e = new Event;
    $e->{srcip} = inet_ntoa($ip);
    $e->{srcport} = $port;
    $e->{timestamp} = $^T;
    $e->{protocol} = 'snmp';

    my($version, $community, $ent, $agent, $gen, $spec, $dt, $bindings)
        = decode_by_template ($buf, "%{%i%s%*{%O%A%i%i%u%{%@",
                              SNMP_Session::trap_request()
                              );

    my ($binding, $oid, $value);
    while ($bindings ne '') {
	($binding,$bindings) = &decode_sequence ($bindings);
	($oid, $value) = decode_by_template ($binding, "%O%@");
	# print "B: ", BER::pretty_oid($oid)," => ", pretty_print($value),"\n";
	$e->{snmp}->{oids}->{ BER::pretty_oid($oid) } = pretty_print($value);

	# cisco sends these (ifno is required):
	$ifno = $e->{snmp}->{ifno} = pretty_print($value) if BER::pretty_oid($oid) =~ /^1.3.6.1.2.1.2.2.1.1/;
    }

    $e->{text} = $gentrap[ $gen ];
    
    # some classification

    if( !$MO::alias{$e->{srcip}} ){
	::error('WARN', "SNMP TRAP '$e->{text}' rcvd from unknown device ($e->{srcip}) (Community: $community)\n");
	return;
    }

    if( $gen == 0 || $gen == 1 ){
	# device reboot
	$e->{object}   = $MO::alias{$e->{srcip}};
	$e->{objtype}  = $e->{object}->{type};
	$e->{class}    = 'event';
	$e->{subclass} = 'reboot';
	$e->{severity} = 8;
	$e->{object}->dev_relearn();
    }
    elsif( $gen == 2 || $gen == 3 ){
	# link up/down
	$e->{object}   = $MO::all{ "$MO::alias{$e->{srcip}}->{caddr}/$ifno" };
	if( ! $e->{object} ){
	    ::error('WARN', "SNMP TRAP for unknown interface ($MO::alias{$e->{srcip}}->{caddr}/$ifno)\n");
	    return;
	}
	$e->{objtype}  = 'intrfc';
	$e->{class}    = 'status';
	if( $gen == 2 ){
	    $e->{subclass} = 'down';
	    $e->{severity} = 9;
	    $e->{text} = "link down - " . $e->{object}->{ifname};
	}else{
	    $e->{subclass} = 'up';
	    $e->{severity} = 5;
	    $e->{text} = "link up - " . $e->{object}->{ifname};
	}
    }
    elsif( $gen == 4 ){
	$e->{object}   = $MO::alias{$e->{srcip}};
	$e->{objtype}  = $e->{object}->{type};
	$e->{class}    = 'security';
	$e->{subclass} = 'attempt';
	$e->{severity} = 5;

	# cisco sends proprietary where from
	if( $e->{snmp}->{oids}{'1.3.6.1.4.1.9.2.1.5.0'} ){
	    $e->{text}.= " from $e->{snmp}->{oids}{'1.3.6.1.4.1.9.2.1.5.0'}";
	}
    }
    else{
	$e->{object}   = $MO::alias{$e->{srcip}};
	$e->{objtype}  = $e->{object}->{type};
	$e->{class}    = 'info';
	$e->{subclass} = "unknown($gen)";
	$e->{severity} = 5;
	$e->{text}     = "";

	foreach ( keys %{$e->{snmp}->{oids}} ){
	    $e->{text} .= "$_:$e->{snmp}->{oids}->{$_} ";
	}

	foreach ( keys %{$e->{snmp}->{oids}} ){
	    # tcpConnState.laddr.lport.raddr.rport
	    if( /^1.3.6.1.2.1.6.13.1.1\.(?:(?:\d+\.){4})(\d+)\.((?:\d+\.){4})(\d+)/ ){
		my ($port, $addr, $portf) = ($1, $2, $3);
		chop($addr);
		$port = "telnet" if $port == 23;
		$e->{class}    = 'event';
		$e->{subclass} = 'tcp_conn';
		$e->{text}     = "tcp connection ($port) from $addr";
		$e->{severity} = 5;
		last;
	    }

	    # ccmHistoryEventConfigSource
	    if( /^1.3.6.1.4.1.9.9.43.1.1.6.1.4/ ){
		$e->{class}    = 'event';
		$e->{subclass} = 'config';
		$e->{text}     = "cisco configure";
		$e->{severity} = 4;
		$e->{object}->dev_relearn();
		last;
	    }
	    
	    # bgpPeerState
	    if( /^1.3.6.1.2.1.15.3.1.2.(\d+\.\d+\.\d+\.\d+)/ ){
		my($v, $peer);
		$v = $e->{snmp}->{oids}{$_};
		$peer = $1;

		# XXX - ought have a peer object
		$e->{class}    = 'info';
		if( $v == 6 ){
		    $e->{subclass} = "bgp_up";
		    $e->{text}     = "bgp peer $peer up";
		}else{
		    $e->{subclass} = "bgp_down";
		    $e->{text}     = "bgp peer $peer down";
		}
		$e->{severity} = 5;
		last;
	    }

	    # frCircuitState
	    # I'd like to easily match this up to the proper (sub)interface
	    if( /^1.3.6.1.2.1.10.32.2.1.3\.(\d+)\.(\d+)/ ){
		my( $pint, $dlci ) = ($1, $2);
		my $v = $e->{snmp}->{oids}{$_};
		my $int = $e->{object}{intfc}[$1] ? $e->{object}{intfc}[$pint]{ifname} : "INT $pint";
		# NB: 1 = invalid, 2 = active, 3 = inactive
		$e->{class}    = 'info';
		$e->{severity} = 6;
		if( $v == 2 ){
		    $e->{subclass} = "frame_up";
		    $e->{text} = "frame pvc change - $int, DLCI $dlci now active";
		}elsif( $v == 1 ){
		    $e->{subclass} = "frame_down";
		    $e->{text} = "frame pvc change - $int, DLCI $dlci deleted";
		}else{
		    $e->{subclass} = "frame_down";
		    $e->{text} = "frame pvc change - $int, DLCI $dlci now inactive";
		}

		# query device to see which subint this is
		my( $x, $oper, $admin, $state, $ne, $te );

		foreach $x ( @{$e->{object}->{intfc}} ){
		    next unless $x;
		    $state = $e->{object}->{state};
		    ($oper, $admin) = $x->querystatus();
		    
		    if( ( $oper eq "up" &&
			 ( ($state eq "bouncing" && $e->{object}->{substate} eq "down")
			  || $state eq "down" ))
		       || ($oper eq "down" &&
			   ( ($state eq "bouncing" && $e->{object}->{substate} eq "up")
			    || $state eq "up" || $state eq "attn" || $state eq "degraded"))){

			$ne = new Event;

			$ne->{srcip}    = $e->{srcip};
			$ne->{srcport}  = $e->{srcport};
			$ne->{timestamp}= $^T;
			$ne->{protocol} = 'snmp';
			$ne->{class}    = "status";
			$ne->{subclass} = $oper;
			$ne->{object}   = $x;
			$ne->{objtype}  = 'intrfc';
			if( $oper eq "up" ){
			    $ne->{severity} = 5;
			}else{
			    $ne->{severity} = 9;
			}
			$te = $e;  $e  = $ne;  $ne = $te;	# preserve order
			$ne->filter();
		    }
		}
		
	    }
	}


    }
    
    $e->filter();
    
}

1;
